"""Class for reading satellite data for Trollflow based Trollduction
using satpy"""

import logging
import time
from copy import deepcopy

from six.moves.urllib.parse import urlparse

from satpy import Scene
from trollflow.workflow_component import AbstractWorkflowComponent
from trollflow_sat import utils
from trollflow.utils import ordered_load


class SceneLoader(AbstractWorkflowComponent):

    """Creates a scene object from a message and loads the required channels.
    """

    logger = logging.getLogger("SceneLoader")

    def __init__(self):
        super(SceneLoader, self).__init__()
        self.use_lock = False

    def pre_invoke(self):
        """Pre-invoke."""
        pass

    def invoke(self, context):
        """Invoke."""
        # Set locking status, default to False
        self.use_lock = context.get("use_lock", False)
        self.logger.debug("Locking is used in compositor: %s",
                          str(self.use_lock))
        if self.use_lock:
            self.logger.debug("Compositor acquires lock of previous "
                              "worker: %s", str(context["prev_lock"]))
            utils.acquire_lock(context["prev_lock"])

        instruments = context.get("instruments", None)
        if instruments is None:
            utils.release_locks([context["lock"], context["prev_lock"]],
                                log=self.logger.error,
                                log_msg="No instruments configured!")
            return

        readers = context.get("readers", None)

        with open(context["product_list"], "r") as fid:
            product_config = ordered_load(fid)
        msg = deepcopy(context['content'])
        for key, val in context.items():
            if key.startswith('ignore_') and val is True:
                msg.data.pop(key[7:], None)

        # Rename the instrument in the message if an alias is given for it
        instrument_aliases = context.get("instrument_aliases", {})
        if instrument_aliases:
            orig_sensor = msg.data['sensor']
            if isinstance(orig_sensor, list):
                orig_sensor = orig_sensor[0]
            sensor = instrument_aliases.get(orig_sensor, orig_sensor)
            if sensor != orig_sensor:
                msg.data['sensor'] = sensor
                self.logger.info(
                    "Adjusted message instrument name from %s to %s",
                orig_sensor, sensor)

        global_data = self.create_scene_from_message(msg, instruments,
                                                     readers=readers)
        if global_data is None:
            utils.release_locks([context["lock"], context["prev_lock"]],
                                log=self.logger.info,
                                log_msg="Unable to create Scene, " +
                                "skipping data")
            return

        monitor_topic = context.get("monitor_topic", None)
        if monitor_topic is not None:
            nameservers = context.get("nameservers", None)
            port = context.get("port", 0)
            service = context.get("service", None)
            monitor_metadata = utils.get_monitor_metadata(msg.data,
                                                          status="start",
                                                          service=service)
            utils.send_message(monitor_topic,
                               "monitor",
                               monitor_metadata,
                               nameservers=nameservers, port=port)

        # TODO: add usage of external calibration coefficients
        # use_extern_calib = product_config["common"].get("use_extern_calib",
        #                                                 "False")

        process_by_area = product_config["common"].get("process_by_area",
                                                       True)
        # Set lock if locking is used
        if self.use_lock:
            self.logger.debug("Compositor acquires own lock %s",
                              str(context["lock"]))
            utils.acquire_lock(context["lock"])

        for area_id in product_config["product_list"]:
            extra_metadata = {}

            # Check if the data was collected for specific area
            if "collection_area_id" in msg.data:
                if area_id != msg.data["collection_area_id"]:
                    utils.release_locks([context["lock"]],
                                        log=self.logger.debug,
                                        log_msg="Collection not for this " +
                                        "area, skipping")
                    continue

            # Load and unload composites for this area
            composites = self.load_composites(global_data, product_config,
                                              area_id)

            extra_metadata['products'] = composites
            extra_metadata['area_id'] = area_id
            context["output_queue"].put({'scene': global_data,
                                         'extra_metadata': extra_metadata})
            if process_by_area:
                context["output_queue"].put(None)


        # Add "terminator" to the queue to trigger computations for
        # this global scene, if not already done
        if not process_by_area:
            context["output_queue"].put(None)

        if utils.release_locks([context["lock"]]):
            self.logger.debug("Compositor releases own lock %s",
                              str(context["lock"]))
            # Wait 1 second to ensure next worker has time to acquire the
            # lock
            time.sleep(1)

        del global_data
        global_data = None

        # Wait until the lock has been released downstream
        if self.use_lock:
            utils.acquire_lock(context["lock"])
            utils.release_locks([context["lock"]])

        if monitor_topic is not None:
            monitor_metadata = utils.get_monitor_metadata(msg.data,
                                                          status="completed",
                                                          service=service)
            utils.send_message(monitor_topic,
                               "monitor",
                               monitor_metadata,
                               nameservers=nameservers,
                               port=port)

        # After all the items have been processed, release the lock for
        # the previous step
        utils.release_locks([context["prev_lock"]], log=self.logger.debug,
                            log_msg="Compositor releases lock of previous "
                            "worker")

    def post_invoke(self):
        """Post-invoke"""
        pass

    def create_scene_from_message(self, msg, instruments, readers=None):
        """Parse the message *msg* and return a corresponding MPOP scene.
        """
        if msg.type in ["file", 'collection', 'dataset']:
            return self.create_scene_from_mda(msg.data, msg.type, instruments,
                                              readers=readers)

    def create_scene_from_mda(self, mda, msg_type, instruments, readers=None):
        """Read the metadata *mda* and return a corresponding MPOP scene.
        """
        if isinstance(mda['sensor'], (list, tuple, set)):
            sensor = mda['sensor'][0]
        else:
            sensor = mda['sensor']

        if sensor not in instruments:
            self.logger.debug("Unknown sensor, skipping data.")
            return None

        filenames = []
        if msg_type == "dataset":
            for dset in mda["dataset"]:
                filenames.append(urlparse(dset["uri"]).path)
        elif msg_type == "collection":
            for col in mda['collection']:
                if 'dataset' in col:
                    for dset in col['dataset']:
                        filenames.append(urlparse(dset["uri"]).path)
                else:
                    filenames.append(urlparse(col["uri"]).path)
        else:
            filenames.append(urlparse(mda['uri']).path)

        # Create satellite scene

        # There can be several readers configured for one instance, so
        # try which matches.  If there's a reader specified in the
        # incoming message, use that.
        readers = mda.get('reader') or readers
        if not isinstance(readers, list):
            readers = [readers]

        global_data = None
        for reader in readers:
            try:
                self.logger.debug("Trying reader %s", reader)
                global_data = Scene(filenames=filenames, reader=reader)
                self.logger.debug("Reader selected.")
                break
            except ValueError:
                continue

        if global_data is not None:
            self.logger.debug("SCENE: %s", str(global_data.attrs))
            global_data.attrs.update(mda)
        else:
            self.logger.warning("Scene not created.")

        return global_data

    def load_composites(self, global_data, product_config, area_id):
        """Get a set of composites for an area"""
        all_composites = \
            set(product_config["product_list"][area_id]['products'].keys())

        # Check solar elevations and remove those composites that
        # are outside of their specified ranges
        composites = set()

        start_time = global_data.attrs['start_time']

        # Check for Sun zenith angle limits
        for composite in all_composites:
            if utils.bad_sunzen_range(
                    product_config,
                    area_id, composite,
                    start_time):
                self.logger.info("Removing composite '%s'; out of "
                                 "valid solar angle range", composite)
            else:
                composites.add(composite)

        # Unload possible pre-existing composites that are not used
        prev_reqs = {itm.name for itm in global_data.datasets}
        reqs_to_unload = prev_reqs - composites
        if len(reqs_to_unload) > 0:
            self.logger.debug("Unloading unnecessary channels: %s",
                              str(sorted(reqs_to_unload)))
            global_data.unload(list(reqs_to_unload))

        self.logger.info("Loading required data for area %s: %s",
                         area_id,
                         ', '.join(sorted(composites)))
        global_data.load(composites)



"""Classes for generating image composites for Trollflow based
Trollduction using satpy"""

import logging
import time

from trollflow.workflow_component import AbstractWorkflowComponent
from trollflow_sat import utils
from trollflow.utils import ordered_load
try:
    from trollsched.satpass import Pass
except ImportError:
    Pass = None


class Resampler(AbstractWorkflowComponent):

    """Creates resampled local area scenes."""

    logger = logging.getLogger("Resampler")

    def __init__(self):
        super(Resampler, self).__init__()
        self.use_lock = False

    def pre_invoke(self):
        """Pre-invoke"""
        pass

    def invoke(self, context):
        """Invoke"""
        # Set locking status, default to False
        self.use_lock = context.get("use_lock", False)
        self.logger.debug("Locking is used in resampler: %s",
                          str(self.use_lock))
        if self.use_lock:
            self.logger.debug("Compositor acquires lock of previous "
                              "worker: %s", str(context["prev_lock"]))
            utils.acquire_lock(context["prev_lock"])

        # Check for terminator
        if context["content"] is None:
            context["output_queue"].put(None)
        else:
            # Process the scene
            self._process(context)

        # After all the items have been processed, release the lock for
        # the previous step
        utils.release_locks([context["prev_lock"]], log=self.logger.debug,
                            log_msg="Resampler releses lock of previous " +
                            "worker: %s" % str(context["prev_lock"]))

    def _process(self, context):
        """Process a context."""

        glbl = context["content"]["scene"]
        extra_metadata = context["content"]["extra_metadata"]

        with open(context["product_list"], "r") as fid:
            product_config = ordered_load(fid)

        # Handle config options
        kwargs = {}

        kwargs['mask_area'] = context.get('mask_area', True)
        self.logger.debug("Setting area masking to %s",
                          str(kwargs['mask_area']))

        kwargs['nprocs'] = context.get('nprocs', 1)
        self.logger.debug("Using %d CPUs for resampling.", kwargs['nprocs'])

        kwargs['resampler'] = context.get('resampler', "nearest")
        self.logger.debug("Using resampling method: '%s'.",
                          kwargs['resampler'])

        try:
            kwargs['cache_dir'] = context['cache_dir']
            self.logger.debug("Setting projection cache dir to %s",
                              kwargs['cache_dir'])
        except (AttributeError, KeyError):
            pass

        kwargs['reduce_data'] = context.get('reduce_data', True)
        self.logger.debug("Reduce data: %s", str(kwargs['reduce_data']))

        prod_list = product_config["product_list"]

        # Overpass for coverage calculations
        scn_metadata = glbl.attrs
        if product_config['common'].get('coverage_check', True) and Pass:
            instrument = scn_metadata['sensor']
            if isinstance(instrument, (list, tuple)):
                instrument = instrument[0]
            overpass = Pass(scn_metadata['platform_name'],
                            scn_metadata['start_time'],
                            scn_metadata['end_time'],
                            instrument=instrument)
        else:
            overpass = None

        # Get the area ID from metadata dict
        area_id = extra_metadata['area_id']

        # Check for area coverage
        if overpass is not None:
            min_coverage = prod_list[area_id].get("min_coverage", 0.0)
            if not utils.covers(overpass, area_id, min_coverage,
                                self.logger):
                return

        kwargs['radius_of_influence'] = None
        try:
            area_config = product_config["product_list"][area_id]
            kwargs['radius_of_influence'] = \
                area_config.get("srch_radius", context["radius"])
        except (AttributeError, KeyError):
            kwargs['radius_of_influence'] = 10000.

        if kwargs['radius_of_influence'] is None:
            self.logger.debug("Using default search radius.")
        else:
            self.logger.debug("Using search radius %d meters.",
                              int(kwargs['radius_of_influence']))
        # Set lock if locking is used
        if self.use_lock:
            self.logger.debug("Resampler acquires own lock %s",
                              str(context["lock"]))
            utils.acquire_lock(context["lock"])

        if area_id == "satproj":
            self.logger.info("Using satellite projection")
            lcl = glbl
        else:
            metadata = glbl.attrs
            self.logger.info("Resampling time slot %s to area %s",
                             metadata["start_time"], area_id)
            lcl = glbl.resample(area_id, **kwargs)

        # Add area ID to the scene attributes so everything needed
        # in filename composing is in the same dictionary
        lcl.attrs["area_id"] = area_id

        metadata = extra_metadata.copy()
        metadata["product_config"] = product_config
        metadata["products"] = prod_list[area_id]['products']

        self.logger.debug("Inserting lcl (area: %s, start_time: %s) "
                          "to writer's queue",
                          area_id, str(scn_metadata["start_time"]))
        context["output_queue"].put({'scene': lcl,
                                     'extra_metadata': metadata})

        if utils.release_locks([context["lock"]]):
            self.logger.debug("Resampler releases own lock %s",
                              str(context["lock"]))
            # Wait 1 second to ensure next worker has time to acquire the
            # lock
            time.sleep(1)

        # Wait until the lock has been released downstream
        if self.use_lock:
            utils.acquire_lock(context["lock"])
            utils.release_locks([context["lock"]])

        del lcl
        lcl = None

    def post_invoke(self):
        """Post-invoke"""
        pass




import logging
import os.path
from six.moves.queue import Empty as queue_empty
import time
from threading import Thread
from satpy.writers import compute_writer_results

from posttroll.message import Message
from posttroll.publisher import Publish
from trollflow_sat import utils
from trollsift import compose


class DataWriterContainer(object):

    '''Container for DataWriter instance
    '''

    logger = logging.getLogger("DataWriterContainer")
    _prev_lock = None

    def __init__(self, topic=None, port=0, nameservers=None,
                 save_settings=None, use_lock=False,
                 publish_vars=None):
        # store parameters for later writer restarts
        self.topic = topic
        self._input_queue = None
        self.output_queue = None
        self.thread = None
        self.use_lock = use_lock
        self._save_settings = save_settings or {}
        self._topic = topic
        self._port = port
        self._nameservers = nameservers
        self._publish_vars = publish_vars
        self._init_writer()

    def _init_writer(self):
        # Create a Writer instance
        self.writer = DataWriter(queue=self.input_queue,
                                 save_settings=self._save_settings,
                                 topic=self._topic,
                                 port=self._port,
                                 nameservers=self._nameservers,
                                 publish_vars=self._publish_vars,
                                 prev_lock=self._prev_lock)
        # Start Writer instance into a new daemonized thread.
        self.thread = Thread(target=self.writer.run)
        self.thread.setDaemon(True)
        self.thread.start()

    @property
    def input_queue(self):
        """Property writer"""
        return self._input_queue

    @input_queue.setter
    def input_queue(self, queue):
        """Set writer queue"""
        self._input_queue = queue
        self.writer.queue = queue

    @property
    def prev_lock(self):
        """Property writer"""
        return self._prev_lock

    @prev_lock.setter
    def prev_lock(self, lock):
        """Set lock of the previous worker"""
        if self.use_lock:
            self._prev_lock = lock
            self.writer.prev_lock = lock

    def __setstate__(self, state):
        self.__init__(**state)

    def restart(self):
        '''Restart writer after configuration update.
        '''
        if self.writer is not None:
            if self.writer.loop:
                self.stop()
        self._init_writer()

    def stop(self):
        '''Stop writer.'''
        self.logger.debug("Stopping writer.")
        self.writer.stop()
        if self.thread is not None:
            self.thread.join()
        self.logger.debug("Writer stopped.")
        self.thread = None

    def is_alive(self):
        """Return the thread status"""
        return self.thread.is_alive()


class DataWriter(Thread):

    """Writes data to disk.
    """

    logger = logging.getLogger("DataWriter")

    def __init__(self, queue=None, save_settings=None,
                 topic=None, port=0, nameservers=None, prev_lock=None,
                 publish_vars=None):
        Thread.__init__(self)
        self.queue = queue
        self._loop = False
        self._save_settings = save_settings
        self._port = port
        if nameservers is None:
            self._nameservers = []
        else:
            if type(nameservers) not in (list, tuple, set):
                nameservers = [nameservers, ]
            self._nameservers = nameservers
        self._topic = topic
        self.prev_lock = prev_lock
        self._publish_vars = publish_vars or {}
        self.data = []
        self.messages = []
        self.pub = None

    def run(self):
        """Run the thread."""
        self._loop = True
        # Get save settings
        kwargs = self._save_settings.copy()

        # Initialize publisher context
        with Publish("l2producer", port=self._port,
                     nameservers=self._nameservers) as self.pub:

            while self._loop:
                if self.queue is not None:
                    try:
                        data = self.queue.get(True, 1)
                        if self.prev_lock is not None:
                            self.logger.debug("Writer acquires lock of "
                                              "previous worker: %s",
                                              str(self.prev_lock))
                            utils.acquire_lock(self.prev_lock)
                        self.queue.task_done()
                    except queue_empty:
                        continue
                    try:
                        if data is None:
                            self._compute()
                            self.data = []
                            self.messages = []
                        else:
                            self._process(data, **kwargs)
                    except Exception:
                        self.logger.exception("Something went wrong when writing.")

                    # After all the items have been processed, release the
                    # lock for the previous worker

                    if self.prev_lock is not None:
                        utils.release_locks([self.prev_lock],
                                            self.logger.debug,
                                            "Writer releses lock of "
                                            "previous worker: %s" %
                                            str(self.prev_lock))
                else:
                    time.sleep(1)

    def _compute(self):
        """Compute the data."""
        if self.data:
            self.logger.info("Processing and saving all data")
            compute_writer_results(self.data)
            if 'overviews' in self._save_settings:
                self._add_overviews()
            self._send_messages()

    def _add_overviews(self):
        """Add overviews (reduced resolution versions of image data) to the
        files.
        """
        fnames = [msg.data['uri'] for msg in self.messages]
        overviews = self._save_settings['overviews']
        utils.add_overviews(fnames, overviews, logger=self.logger)

    def _send_messages(self):
        """Send currently collected messages about completed datasets."""
        for msg in self.messages:
            self.pub.send(str(msg))
            self.logger.debug("Sent message: %s", str(msg))

    def _process(self, data, **kwargs):
        """Process the incoming data lazily"""
        lcl = data['scene']
        extra_metadata = data['extra_metadata']
        product_config = extra_metadata["product_config"]
        products = extra_metadata["products"]

        scn_metadata = lcl.attrs.copy()

        # Available composite names
        composite_names = [dset.name for dset in lcl.keys()]

        # Save all products in a delayed way
        for prod in products:
            # Skip the removed composites
            if prod not in composite_names:
                continue

            # Create output filenames for this product
            fnames, productname = \
                utils.create_fnames(scn_metadata,
                                    product_config,
                                    prod)
            # Some of the files might have specific format settings,
            # so read them from the config.  Use SatPy defaults if
            # nothing is given.
            fmts = utils.get_format_settings(product_config, prod,
                                             scn_metadata["area_id"])

            # Read writer specific kwargs 
            writer_kwargs = utils.read_writer_config(product_config, products[prod], prod, scn_metadata)
            kwargs.update(writer_kwargs)


            # Create delayed writer objects and messages
            for j, fname in enumerate(fnames):
                try:
                    dset = lcl.save_datasets(datasets=[prod],
                                             filename=fname,
                                             writer=fmts[j]['writer'],
                                             fill_value=fmts[j]['fill_value'],
                                             compute=False,
                                             **kwargs)
                except Exception:
                    self.logger.exception("Something went wrong when saving %s to %s.", prod, fname)
                    continue
                self.data.append(dset)

                # Create message for this file
                self._create_message(lcl[prod].attrs.get("area"),
                                     fname, scn_metadata, productname)

    def _create_message(self, area, fname, scn_metadata, productname):
        """Create a message and add it to self.messages"""
        # No messaging without a topic
        if self._topic is None:
            return

        try:
            area_data = {"name": area.name,
                         "area_id": area.area_id,
                         "proj_id": area.proj_id,
                         "proj4": area.proj4_string,
                         "shape": (area.x_size,
                                   area.y_size)
                         }
        except AttributeError:
            area_data = None

        # Create message metadata dictionary
        to_send = \
            utils.select_dict_items(scn_metadata,
                                    self._publish_vars)

        to_send_fix = {"nominal_time": scn_metadata["start_time"],
                       "uid": os.path.basename(fname),
                       "uri": os.path.abspath(fname),
                       "area": area_data,
                       "productname": productname
                       }
        to_send.update(to_send_fix)

        topic = self._topic
        # Compose the topic with area information
        if area_data is not None:
            tmp = to_send.copy()
            del tmp["area"]
            area_data.update(tmp)
            topic = compose(topic, area_data)
        else:
            topic = compose(topic,
                            {'area_id': 'satproj'})

        # Create message
        msg = Message(topic, "file", to_send)
        self.messages.append(msg)

    def stop(self):
        """Stop writer."""
        self._loop = False

    @property
    def loop(self):
        """Property loop"""
        return self._loop



import logging
import os.path

from posttroll.message import Message
from posttroll.publisher import Publish
from trollflow.utils import acquire_lock as trollflow_acquire_lock
from trollflow.utils import release_lock
from trollsift import compose
from trollsift.parser import get_convert_dict

from satpy.resample import get_area_def

try:
    from pyorbital import astronomy
except ImportError:
    astronomy = None

try:
    import dpath.util
    DPATH_AVAILABLE = True
except ImportError:
    DPATH_AVAILABLE = False

PATTERN = "{time:%Y%m%d_%H%M}_{platform_name}_{areaname}_{productname}.png"

FORMAT_DEFAULTS = {'writer': None,
                   'format': None,
                   'fill_value': None}

LOGGER = logging.getLogger(__name__)


def create_fnames(info, product_config, prod_id):
    """Create filename for product *prod*"""
    area_id = info["area_id"]

    # List of products
    products = product_config["product_list"][area_id]["products"]

    # Get area name
    info["areaname"] = product_config["product_list"][area_id]["areaname"]

    # Find output directory
    output_dir = products[prod_id].get("output_dir", "")
    if output_dir == "":
        output_dir = product_config["common"].get("output_dir", "")

    if output_dir == "":
        LOGGER.warning("No output directory specified, "
                       "saving to current directory!")

    # Find filename pattern
    pattern = products[prod_id].get("fname_pattern", "")
    if pattern == "":
        pattern = product_config["common"].get("fname_pattern", "")

    if pattern == "":
        LOGGER.warning("No pattern was given, using built-in default: %s",
                       PATTERN)
        pattern = PATTERN

    # Join output dir and filename pattern
    pattern = os.path.join(output_dir, pattern)

    # Find output formats
    formats = products[prod_id].get("formats", None)
    if formats is None:
        formats = product_config["common"].get("formats", [FORMAT_DEFAULTS])

    prod_name = products[prod_id]["productname"]
    info["productname"] = prod_name

    # Find the name of the available 'nominal_time'
    time_name = None
    for key in info:
        if "time" in key and "end" not in key and "proc" not in key:
            time_name = key
            LOGGER.debug("metadata time name is '%s'", time_name)
            break

    if time_name is None and "time" in pattern:
        return None, None

    # Adjust filename pattern so that time_name is present.
    # Get parse definitions and try to figure out if there's
    # an item for time
    convert_dict = get_convert_dict(pattern)
    for key, val in convert_dict.items():
        if ("time" in key or "%" in val) and \
           "end" not in key and key != time_name:
            LOGGER.debug("Updating pattern from '%s' ...", pattern)

            while '{' + key in pattern:
                pattern = pattern.replace('{' + key,
                                          '{' + time_name)
            LOGGER.debug("... to '%s'", pattern)

    fnames = []
    for fmt in formats:
        # Default to TIFF
        if pattern.endswith("{format}") and fmt["format"] is None:
            fmt["format"] = "tif"
        info["format"] = fmt["format"]
        # Ensure non-unicode filename
        fnames.append(str(compose(pattern, info)))

    return (fnames, prod_name)


def get_format_settings(product_config, prod_id, area_id):
    """Get all the format settings for this product"""
    products = product_config["product_list"][area_id]["products"]
    formats = products[prod_id].get("formats", [{}])

    settings = []
    for fmt in formats:
        tmp = {}
        for key in FORMAT_DEFAULTS.keys():
            val = fmt.get(key, FORMAT_DEFAULTS[key])
            tmp[key] = val
        settings.append(tmp)

    return settings


def bad_sunzen_range(product_config, area_id, composite, start_time):
    """Check if Sun zenith angle is valid at the configured location.
    SatPy version.
    """
    product_conf = \
        product_config["product_list"][area_id]["products"][composite]

    if ("sunzen_night_minimum" not in product_conf and
            "sunzen_day_maximum" not in product_conf):
        return False

    if astronomy is None:
        LOGGER.warning("Pyorbital not installed, unable to calculate "
                       "Sun zenith angles!")
        return False

    if "sunzen_lon" not in product_conf and "sunzen_lat" not in product_conf:
        LOGGER.warning("No 'sunzen_lon' or 'sunzen_lat' configured, "
                       "can\'t check Sun elevation.")
        return False

    lon = product_conf["sunzen_lon"]
    lat = product_conf["sunzen_lat"]
    sunzen = astronomy.sun_zenith_angle(start_time, lon, lat)
    LOGGER.debug("Sun zenith angle is %.2f degrees", sunzen)

    try:
        limit = product_conf["sunzen_night_minimum"]
        if sunzen < limit:
            return True
        else:
            return False
    except KeyError:
        pass

    limit = product_conf["sunzen_day_maximum"]
    if sunzen > limit:
        return True
    else:
        return False


def send_message(topic, msg_type, msg_data, nameservers=None, port=0):
    """Send monitoring message"""
    if nameservers is None:
        nameservers = []
    if not isinstance(nameservers, list):
        nameservers = [nameservers]

    with Publish("trollflow-sat", port=port,
                 nameservers=nameservers) as pub:
        msg = Message(topic, msg_type, msg_data)
        pub.send(str(msg))


def _get_data_time_from_message_data(msg_data):
    """Get data timestamp from message data"""
    if "nominal_time" in msg_data:
        data_time = msg_data["nominal_time"]
    elif "start_time" in msg_data:
        data_time = msg_data["start_time"]
    elif "time" in msg_data:
        data_time = msg_data["time"]
    else:
        data_time = None

    return data_time


def _get_orbit_number_from_message_data(msg_data):
    """Get orbit number from message data"""
    if "orbit_number" in msg_data:
        return msg_data["orbit_number"]
    else:
        return None


def get_monitor_metadata(msg, status=None, service=None):
    """Collect metadata for monitoring message"""
    data = {"message_time": msg.time,
            "data_time": _get_data_time_from_message_data(msg.data),
            "platform_name": msg.data["platform_name"],
            "sensor": msg.data["sensor"],
            "orbit_number": _get_orbit_number_from_message_data(msg.data),
            "status": status,
            "service": service}

    return data


def release_locks(locks, log=None, log_msg=None):
    """Release locks and optionnally send log message to *log* function"""
    if not isinstance(locks, list):
        locks = [locks]
    if log is not None and log_msg is not None:
        log(log_msg)
    ret_vals = []
    for lock in locks:
        ret_vals.append(release_lock(lock))

    return max(ret_vals)


def acquire_lock(lock):
    """Acquire the given lock"""
    return trollflow_acquire_lock(lock)


def covers(overpass, area_name, min_coverage, logger):
    try:
        area_def = get_area_def(area_name)
        if min_coverage == 0 or overpass is None:
            return True
        min_coverage /= 100.0
        coverage = overpass.area_coverage(area_def)
        if coverage <= min_coverage:
            logger.info("Coverage too small %.1f%% (out of %.1f%%) "
                        "with %s",
                        coverage * 100, min_coverage * 100,
                        area_name)
            return False
        else:
            logger.info("Coverage %.1f%% with %s",
                        coverage * 100, area_name)

    except AttributeError:
        logger.warning("Can't compute area coverage with %s!",
                       area_name)
    return True


def select_dict_items(src_dict, selection):
    """Creates a new dictionary containing elements listed in selection"""
    to_send = dict(src_dict) if '*' in selection else {}

    for dest_key in selection:
        if dest_key != '*':
            if isinstance(selection, dict):
                val = selection[dest_key]
                if '/' in val:
                    if not DPATH_AVAILABLE:
                        LOGGER.error("path expressions in publish_var but no dpath available")
                    else:
                        # use dpath for path expressions
                        info_without_empty_keys = \
                            {k: v for k, v in src_dict.items() if k}
                        if '*' in val:
                            # returns list
                            to_send[dest_key] = \
                                dpath.util.values(info_without_empty_keys, val)
                        else:
                            # returns single value
                            to_send[dest_key] = \
                                dpath.util.get(info_without_empty_keys, val)
                else:
                    to_send[dest_key] = src_dict.get(val)
            else:
                to_send[dest_key] = dest_key
    return to_send


def add_overviews(fnames, overviews, logger=None):
    """Add overviews to given files."""
    try:
        import rasterio
        from rasterio.enums import Resampling
    except ImportError:
        if logger is not None:
            logger.error("Can't add overviews, install rasterio")
        return

    if logger is not None:
        logger.info("Adding overviews")

    for fname in fnames:
        try:
            with rasterio.open(fname, 'r+') as dst:
                dst.build_overviews(overviews, Resampling.average)
                dst.update_tags(ns='rio_overview', resampling='average')
        except rasterio.RasterioIOError:
            pass


def read_writer_config(product_config, single_product_config, product, scn_metadata):
    """Execute writer config callback to add extra kwargs to the writer
        :Parameters:
            product_config: dict
                products config parameters
            single_product_config: dict
                config parameters of the current product
            product: str
                current product name
            scn_metadata: dict
                extra metadata

        :Returns:
            extra kwargs: dict
    """

    if "writer_config" in product_config['common']:
        func = product_config['common']['writer_config']['method']
        config_fname = product_config['common']['writer_config']['config_filename']
        return func(config_fname, product, single_product_config, scn_metadata)
    else:
        return dict()
